
import os
import sys
import json
import optparse

from .volume import VMVolume, VolumeSet

ENV_PREFIX = 'VMAKEDK'
VERSION = '0.2.1'


def main():
    parser = optparse.OptionParser(
        usage=('%prog <command> [options]\n'
               '\n'
               'Commands:\n'
               ' open     Create disks and mount virtual filesystem\n'
               ' reopen   (re)Mount previously created disks\n'
               ' fstab    Output the fstab to stdout'),
        version=VERSION,
        conflict_handler='resolve',
    )

    parser.add_option(
        '-h', '--help',
        action='help',
        help='Print help text and exit')
    parser.add_option(
        '-v', '--version',
        action='version',
        help='Print version and exit')
    parser.add_option(
        '-d', '--dry-run',
        action='store_true', default=False,
        help='Only show commands that would have been executed')
    parser.add_option(
        '-t', '--store',
        dest='store',
        default=os.environ.get(ENV_PREFIX + 'STORE', None),
        help='Path to vmdk storage (or %sSTORE from env)' % (ENV_PREFIX,))
    parser.add_option(
        '-r', '--root',
        dest='root',
        default=os.environ.get(ENV_PREFIX + 'ROOT', None),
        help='Path to vm (ch)root (or %sROOT from env)' % (ENV_PREFIX,))
    parser.add_option(
        '-l', '--layout',
        dest='layout',
        default=os.environ.get(ENV_PREFIX + 'LAYOUT', None),
        help='Path to layout json (or %sLAYOUT from env)' % (ENV_PREFIX,))
    parser.add_option(
        '-c', '--create-type',
        dest='create_type',
        default=os.environ.get(ENV_PREFIX + 'CREATE_TYPE', 'monolithicFlat'),
        help='vmdk create_type (or %sCREATE_TYPE from env) Must be either'
             ' "monolithicFlat" or "vmfs"' % (ENV_PREFIX,))

    opts, args = parser.parse_args(sys.argv[1:])
    if not opts.store:
        parser.error('must provide a directory for vmdk storage')
    if not opts.root:
        parser.error('must provide a directory for vm root/sandbox')
    if not opts.layout:
        parser.error('must provide a file with the vm disk layout json')
    if not args:
        parser.error('must provide a command')
    if len(args) > 1:
        parser.error('too many commands')
    command = args[0]

    if opts.dry_run:
        patch_dry_run()

    with open(opts.layout, 'r') as fd:
        layout = json.load(fd)
    volume_set = VolumeSet(
        volumes=(VMVolume.from_dict(vol) for vol in layout['volumes'])
    )


    if command == 'open' or command == 'reopen':
        if command == 'open':
            volume_set.create_data(opts.store, opts.create_type)
        volume_set.loop_data(opts.store)
        if command == 'open':
            volume_set.mkfs()
        volume_set.mount(opts.root)

    elif command == 'fstab':
        print volume_set.fstab()
        sys.exit(0)

    else:
        parser.error('unknown command %s' % (command,))


"""
elif cmd=='close':
    volume_set.umount(SANDBOX)
    volume_set.unloop_data()
"""

def patch_dry_run():
    from .runner import DryRunner
    import volume
    import loopback
    loopback.Loopback.dry_run = True
    volume.runner = DryRunner()





SCSI_DISK_TEMPLATE = ('scsi{{bus}}:{{id}}.present = "TRUE"\n'
                      'scsi{{bus}}:{{id}}.fileName = "{{filename}}"\n'
                      'scsi{{bus}}:{{id}}.deviceType = "scsi-hardDisk"\n'
                      'scsi{{bus}}:{{id}}.redo = ""\n')

SCSI_BUS_TEMPLATE = ('scsi{{bus}}.present = "TRUE"\n'
                     'scsi{{bus}}.virtualDev = "pvscsi"\n')

scsi_ids = (num for num in range(0, 16) if num != 7)





import os
from itertools import chain

from .vmdk import make_vmdk
from .runner import RealRunner
from .loopback import Loopback, LoopNotFound
from .procmount import ProcMounts
from .vfstree import  VfsTree


runner = RealRunner()


class VolumeSet(object):

    def __init__(self, volumes=None):
        self.itree = VfsTree()
        self.nontree = []
        if volumes:
            for vol in volumes:
                self.add(vol)

    def add(self, vol):
        nodes, vol = vol.tree_ref()
        if nodes:
            self.itree.follow_and_set(nodes, vol)
        else:
            self.nontree.append(vol)

    def detect(self, store, sandbox):
        for vol in self.itree.sorted():
            vol.detect(store, sandbox)
            if vol.loop:
                print 'existing volume found on', vol.loop, ', ', 'mounted' if vol.up else 'not mounted'

    def fstab(self):
        node_entries = (vol.fstab_entry() for vol in self.itree.sorted() if vol.include_in_fstab)
        none_entries = (vol.fstab_entry() for vol in self.nontree if vol.include_in_fstab)
        all_entries = chain(node_entries, none_entries)
        return "\n".join(all_entries)

    def create_data(self, store, create_type):
        for vol in chain(self.itree.sorted(), self.nontree):
            vol.create_data(store)
            vol.create_vmdk(store, create_type)

    def loop_data(self, store):
        for vol in chain(self.itree.sorted(), self.nontree):
            vol.loop_data(store)

    def mkfs(self):
        for vol in chain(self.itree.sorted(), self.nontree):
            vol.mkfs()

    def mount(self, sandbox):
        for vol in self.itree.sorted():
            vol.mount(sandbox)

    def umount(self, sandbox):
        for vol in reversed(list(self.itree.sorted())):
            vol.umount(sandbox)

    def unloop_data(self):
        for vol in chain(self.itree.sorted(), self.nontree):
            vol.unloop_data()


class VMVolume(object):

    dump = 0
    fstype = 'raw'

    def loop_data(self, dir_):
        fn = os.path.join(dir_, self.dataname)
        #my_exec("losetup --find %s" % fn)  # safer: add then find
        next_loop_device = Loopback.get_device()
        Loopback.connect(next_loop_device, fn)
        self.loop = next_loop_device

    def unloop_data(self):
        if not self.loop:
            raise Exception("attempted to unloop the unlooped")
        Loopback.disconnect(self.loop)
        self.loop = None

    def __init__(self, label, vmdkname=None, dataname=None, include_in_fstab=None, options=None):
        self.label = label
        self.vmdkname = vmdkname if vmdkname else label+'.vmdk'
        self.dataname = dataname if dataname else label+'.'+self.fstype
        if include_in_fstab is None:
            self.include_in_fstab = True
        else:
            self.include_in_fstab = bool(include_in_fstab)
        if options is None:
            self.options = 'defaults'
        else:
            self.options = options

        self.loop = None
        self.up = False

    def detect(self, store, sandbox):
        data_fn = os.path.join(store, self.dataname)
        sand_fn = os.path.normpath(sandbox + self.mountpoint)

        try:
            self.loop = Loopback.find(data_fn)
        except LoopNotFound:
            self.loop = None

        if self.loop:
            mounts = list(ProcMounts.find(device=self.loop))
            if len(mounts) > 1:
                raise Exception("same device is mounted multiple times")
            elif len(mounts) == 0:
                self.up = False
            else:
                if mounts[0].path == sand_fn:
                    self.up = True
                else:
                    print mounts[0].path
                    print sand_fn
                    raise Exception("unexpected mountpoint for existing loop device")
        else:
            self.up = False


    def mount(self, dir_):
        finaldir = dir_ + self.mountpoint
        runner.mkdir_p(finaldir)
        runner.mountfs(self.loop, finaldir)
        self.up = True

    def umount(self, dir_):
        if not self.up:
            raise Exception("attempt to umount volume that is not up")
        finaldir = dir_ + self.mountpoint
        runner.umountfs(finaldir)
        self.up = False

    def tree_ref(self):
        if hasattr(self, 'mountpoint'):
            return path_to_list(self.mountpoint), self
        else:
            return [], self

    def fstab_entry(self):
        return ' '.join([
            'LABEL='+self.label,
            getattr(self,'mountpoint','none'),
            self.fstype,
            self.options,
            str(self.dump),
            str(self.check),
        ])

    @classmethod
    def from_dict(cls, dict_repr):
        volume_type = dict_repr.pop('type').strip().lower()
        if volume_type == VolumeExt4.volume_type:
            return VolumeExt4(**dict_repr)
        elif volume_type == VolumeXfs.volume_type:
            return VolumeXfs(**dict_repr)
        elif volume_type == VolumeBoot.volume_type:
            return VolumeBoot(**dict_repr)
        elif volume_type == VolumeSwap.volume_type:
            return VolumeSwap(**dict_repr)
        else:
            raise ValueError("Unknown volume type \"%s\"" % volume_type)

    def create_vmdk(self, dir_, create_type):
        fn = os.path.join(dir_, self.vmdkname)
        if os.path.isfile(fn):
            raise Exception('file exists "%s"' % (self.vmdkname,))
        vmdkdata = self.vmdk(create_type)
        with open(fn, 'w') as fd:
            fd.write(vmdkdata)


class VMSparseVolume(VMVolume):
    def __init__(self, label, sectors, vmdkname=None, dataname=None, include_in_fstab=None, options=None):
        super(VMSparseVolume, self).__init__(label, vmdkname=vmdkname, dataname=dataname, include_in_fstab=include_in_fstab, options=options)
        self.sectors = int(sectors)

    def create_data(self, dir_):
        fn = os.path.join(dir_, self.dataname)
        if os.path.isfile(fn):
            raise Exception("file exists")
        runner.create_sparse_file(fn, self.sectors)

    def vmdk(self, create_type):
        return make_vmdk(self.dataname, self.sectors, create_type=create_type)


class VolumeExt4(VMSparseVolume):
    volume_type = 'ext4'
    fstype = 'ext4'
    def __init__(self, label, size, mountpoint, check, options=None, vmdkname=None, dataname=None, include_in_fstab=None):
        super(VolumeExt4, self).__init__(label, decode_size(size), vmdkname=vmdkname, dataname=dataname, include_in_fstab=include_in_fstab, options=options)
        self.mountpoint = os.path.normpath(mountpoint)
        self.check = check

    def mkfs(self):
        if not self.loop:
            raise Exception("volume is not looped")
        #print 'exec:', "mkfs.ext4 -q -F -L %s %s" % (self.label, self.loop)
        runner.passthru(['mkfs.ext4', '-q', '-F', '-L', self.label, self.loop])


class VolumeXfs(VMSparseVolume):
    volume_type = 'xfs'
    fstype = 'xfs'
    def __init__(self, label, size, mountpoint, check, options=None, vmdkname=None, dataname=None, include_in_fstab=None):
        super(VolumeXfs, self).__init__(label, decode_size(size), vmdkname=vmdkname, dataname=dataname, include_in_fstab=include_in_fstab, options=options)
        self.mountpoint = os.path.normpath(mountpoint)
        self.check = check

    def mkfs(self):
        if not self.loop:
            raise Exception("volume is not looped")
        #print 'exec:', "mkfs.ext4 -q -F -L %s %s" % (self.label, self.loop)
        runner.passthru(['mkfs.xfs', '-q', '-f', '-L', self.label, self.loop])


class VolumeSwap(VMSparseVolume):
    volume_type = 'swap'
    fstype = 'swap'
    check = 0
    def __init__(self, label, size, vmdkname=None, dataname=None, include_in_fstab=None):
        super(VolumeSwap, self).__init__(label, decode_size(size), vmdkname=vmdkname, dataname=dataname, include_in_fstab=include_in_fstab, options='defaults')

    def mkfs(self):
        if not self.loop:
            raise Exception("volume is not looped")
        #print 'exec:', "mkswap -f -L %s %s" % (self.label, self.loop)
        runner.passthru(['mkswap', '-f', '-L', self.label, self.loop])

    def mount(self, dir_):
        pass

    def umount(self, dir_):
        pass


class VolumeBoot(VMVolume): # extends ext4?
    volume_type = 'boot'
    fstype = 'ext4'
    def __init__(self, label, source, offset, mountpoint, check, vmdkname=None, dataname=None, include_in_fstab=None, options=None):
        super(VolumeBoot, self).__init__(label, vmdkname=vmdkname, dataname=dataname, include_in_fstab=include_in_fstab, options=options)
        self.source = source
        self.offset = int(offset)
        self.mountpoint = mountpoint
        self.check = check
        #XXX self.sectors = os.stat(source).st_size / 512
        self.sectors = 1024

    def create_data(self, dir_):
        fn = os.path.join(dir_, self.dataname)
        if os.path.isfile(fn):
            raise Exception("file exists")
        runner.copy_file(self.source, fn)

    def loop_data(self, dir_):
        fn = os.path.join(dir_, self.dataname)
        #my_exec("losetup --find %s" % fn)  # safer: add then find
        next_loop_device = Loopback.get_device()
        Loopback.connect(next_loop_device, fn, sectors_to_bytes(self.offset))
        self.loop = next_loop_device

    def mkfs(self):
        if not self.loop:
            raise Exception("volume is not looped")
        #print 'exec:', "mkfs.ext4 -q -F -L %s %s" % (self.label, self.loop)
        runner.passthru(['mkfs.ext4', '-q', '-F', '-L', self.label, self.loop])

    def vmdk(self, create_type):
        return make_vmdk(self.dataname, self.sectors, create_type=create_type)


def path_to_list(path):
    """split a dir path into a list, adding a sentinal value for the root

    Args:
        path: string unix directory pathname e.g. '/usr/bin'

    Returns:
        list of each node in the path, plus a prepended '__root__'

    Raises:
        Exception if the input fails some validity tests
    """
    if path != path.strip():
        raise Exception('whitespace found in path "%s"' % (path,))
    if not path:
        raise Exception('empty path?')
    if path[0] != '/':
        raise Exception('path "%s" has no leading /' % (path,))
    path = '__root__' + path
    return os.path.normpath(path).split('/')


def decode_size(size_string):
    tsz = size_string.strip()
    lasttwo = tsz[-2:]
    amount = int(tsz[0:-2].strip())
    if amount <= 0:
        raise ValueError('bad size')
    if lasttwo.lower() == 'gb':
        return (amount * 1024 * 1024 * 1024) / 512
    elif lasttwo.lower() == 'mb':
        return (amount * 1024 * 1024) / 512
    else:
        raise ValueError('bad size')


def sectors_to_bytes(sectors):
    return sectors * 512



import unittest

from vmakedk.vmdk import sectors_to_geometry

class TestGeometry(unittest.TestCase):
    """test some disk sizes match geometry created with ESXi 6.0
    """

    def test_less_than_1gb(self):
        heads, spt, cylinders = sectors_to_geometry(409600)
        self.assertEquals(heads, 64)
        self.assertEquals(spt, 32)
        self.assertEquals(cylinders, 200)

    def test_greater_than_1gb_and_less_than_2gb(self):
        heads, spt, cylinders = sectors_to_geometry(3276800)
        self.assertEquals(heads, 128)
        self.assertEquals(spt, 32)
        self.assertEquals(cylinders, 800)

    def test_greater_than_2gb(self):
        heads, spt, cylinders = sectors_to_geometry(4915200)
        self.assertEquals(heads, 255)
        self.assertEquals(spt, 63)
        self.assertEquals(cylinders, 305)

    def test_16gb(self):
        heads, spt, cylinders = sectors_to_geometry(33554432)
        self.assertEquals(heads, 255)
        self.assertEquals(spt, 63)
        self.assertEquals(cylinders, 2088)

if __name__ == '__main__':
    unittest.main()





SECTORS_1GB = 1 * 1024 * 1024 * 1024 / 512
SECTORS_2GB = 2 * 1024 * 1024 * 1024 / 512

ADAPTER_TYPE_IDE = 'ide'
ADAPTER_TYPE_BUSLOGIC = 'buslogic'
ADAPTER_TYPE_LSILOGIC = 'lsilogic'

CREATE_TYPE_VMFS = 'vmfs'
CREATE_TYPE_MONOLITHIC_FLAT = 'monolithicFlat'

EXTENT_TYPES = {
    CREATE_TYPE_VMFS: 'VMFS',
    CREATE_TYPE_MONOLITHIC_FLAT: 'FLAT',
}


VMDK_TEMPLATE = ('version=1\n'
                 'encoding="UTF-8"\n'
                 'CID=fffffffe\n'
                 'parentCID=ffffffff\n'
                 'isNativeSnapshot="no"\n'
                 'createType="{{create_type}}"\n'
                 'RW {{sectors}} {{extent_type}} "{{filename}}" 0\n'
                 'ddb.adapterType = "{{adapter_type}}"\n'
                 'ddb.geometry.cylinders = "{{cylinders}}"\n'
                 'ddb.geometry.heads = "{{heads}}"\n'
                 'ddb.geometry.sectors = "{{sectors_per_track}}"\n'
                 'ddb.virtualHWVersion = "6"\n')

def make_vmdk(filename, sectors,
              create_type, adapter_type=ADAPTER_TYPE_LSILOGIC):
    """Render a vmdk descriptor file as a string.

    It is assumed that a single extent covers the entire disk.

    Args:
        filename: filename of the backing storage/extent
        sectors: total disk/extent sectors
        create_type: one of CREATE_TYPE_* constants
        adapter_type: one of ADAPTER_TYPE_* constants

    Returns:
        string containing the vmdk descriptor file data
    """

    heads, spt, cylinders = sectors_to_geometry(sectors)

    return VMDK_TEMPLATE.replace('{{filename}}', filename)\
                        .replace('{{create_type}}', create_type)\
                        .replace('{{extent_type}}', EXTENT_TYPES[create_type])\
                        .replace('{{adapter_type}}', adapter_type)\
                        .replace('{{sectors}}', str(sectors))\
                        .replace('{{cylinders}}', str(int(cylinders)))\
                        .replace('{{sectors_per_track}}', str(int(spt)))\
                        .replace('{{heads}}', str(int(heads)))


def sectors_to_geometry(sectors):
    """Calculate disk geometry based on total sector count

    See VMware kb # 1026266 for details

    Args:
        sectors: disk's total number of sectors

    Returns:
        tuple of (#heads, #sectors-per-track, #cylinders)
    """
    if sectors < SECTORS_1GB:
        return 64, 32, sectors / (64 * 32)
    elif sectors < SECTORS_2GB:
        return 128, 32, sectors / (128 * 32)
    else:
        return 255, 63, sectors / (255 * 63)





import subprocess

class RealRunner(object):
    def mkdir_p(self, path):
        subprocess.check_output(['mkdir', '-p', path])

    def mountfs(self, dev, dir):
        subprocess.check_output(['mount', dev, dir])

    def bindfs(self, src, dir):
        subprocess.check_output(['mount', '-o', 'bind', src, dir])

    def umountfs(self, dir):
        subprocess.check_output(['umount', dir])

    def create_sparse_file(self, filename, sectors):
        subprocess.check_output(['dd', 'of='+filename, 'seek='+str(sectors), 'count=0'])

    def copy_file(self, src, dst):
        subprocess.check_output(['dd', 'if='+src, 'of='+dst])

    def passthru(self, params):
        return subprocess.check_output(params)


class DryRunner(object):
    def mkdir_p(self, path):
        print 'exec: mkdir -p', path

    def mountfs(self, dev, dir):
        print 'exec: mount', dev, dir

    def bindfs(self, src, dir):
        print 'exec: mount -o bind', src, dir

    def umountfs(self, dir):
        print 'exec: umount', dir

    def create_sparse_file(self, filename, sectors):
        print 'exec: dd of=%s bs=512 seek=%d count=0' % (filename, sectors)

    def copy_file(self, src, dst):
        print 'exec: dd if=%s of=%s' % (src, dst)

    def passthru(self, params):
        print 'exec:', ' '.join(params)





import subprocess


class LoopNotFound(Exception):
    pass


class Loopback(object):

    exe = ['losetup']
    fake_dev = 0
    dry_run = False

    @classmethod
    def get_device(cls):
        dev = cls.run(['-f']).strip()
        return dev

    @classmethod
    def connect(cls, loop, filename, offset=None):
        offset_params = ['-o', str(int(offset))] if offset else []
        cls.run(offset_params + [loop, filename])

    @classmethod
    def disconnect(cls, loop):
        cls.run(['-d', loop])

    @classmethod
    def run(cls, params):
        if cls.dry_run:
            print 'running: ', ' '.join(['losetup'] + params)
            if params[0] == '-f':
                cls.fake_dev += 1
                return '/dev/loop' + str(cls.fake_dev)
        else:
            return subprocess.check_output(['losetup'] + params)

    @classmethod
    def find(cls, fn):
        def loopparse(s):
            left,mid,right = s.split(' ')
            dev_part = left[:-1] # trim the ':'
            file_part = right[1:-1]
            return dev_part, file_part

        data = subprocess.check_output(['losetup', '-a']).strip()
        lines = data.split("\n")
        pairs = (loopparse(line) for line in lines)
        for dev, file_ in pairs:
            if file_ == fn:
                return dev
        raise LoopNotFound()




#!/usr/bin/env python
"""
vmakedk module/executable bootstrap

inspired ;) by youtube-dl's layout and __main__

Execute with
# python vmakedk/__main__.py (2.6)
# python -m vmakedk          (2.7)

"""
import sys

if __package__ is None and not hasattr(sys, 'frozen'):
    # this is a direct call to __main__.py
    import os.path
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.append(os.path.dirname(os.path.dirname(path)))

import vmakedk

if __name__ == '__main__':
    vmakedk.main()




class VfsTree(object):
    def __init__(self):
        self.root = VfsNode('')

    def follow_and_set(self, nodelist, inode):
        r = self.root
        for item in nodelist[1:]:
            r = r.add_or_get(item)
        if inode:
            r.set(inode)

    def sorted(self):
        return self.root.sorted()


class VfsNode(object):
    def __init__(self, name, data=None):
        self.name = name
        self.children = []
        self.data = data

    def add_or_get(self, name):
        for child in self.children:
            if child.name == name:
                return child
        new_child = VfsNode(name)
        self.children.append(new_child)
        return new_child

    def set(self, val):
        if self.data:
            raise Exception("item %s already set!" % self.name)
        else:
            self.data = val

    def sorted(self):
        if self.data:
            yield self.data
        for child in self.children:
            for item in child.sorted():
                yield item




from distutils.core import setup

setup(
    name='vmakedk',
    description='Tools for creating virtual machine images',
    long_description=open('README.txt').read(),
    url='https://github.com/benjamin9999/vmakedk/',
    version='0.2.1',
    author='Benjamin Yates',
    author_email='benjamin@rqdq.com',
    packages=['vmakedk', 'vmakedk.test'],
    scripts=['bin/vmakedk'],
    license='Creative Commons Attribution-ShareAlike 3.0 Unported License',
    install_requires=[],
)




class ProcMount(object):
    def __init__(self, device, path, fstype, options):
        self.path = path
        self.fstype = fstype
        self.device = device
        self.options = options

    @classmethod
    def from_procfs_line(cls, line):
        parts = line.split(' ')
        return cls(parts[0], parts[1], parts[2], parts[3])

    def __repr__(self):
        return '<ProcMount %s=%s (%s, %s)>' %\
               (self.path, self.device, self.fstype, self.options)


class ProcMounts(object):
    @classmethod
    def objects(cls):
        with open('/proc/mounts', 'r') as fd:
            for line in fd:
                yield ProcMount.from_procfs_line(line.strip())

    @classmethod
    def find(cls, device=None, path=None, fstype=None):
        for mount in cls.objects():
            match = True
            match &= (path and mount.path == path)
            match &= (device and mount.device == device)
            match &= (fstype and mount.fstype == fstype)
            if match:
                yield mount

